/*
 * Web site error daemon - action handler
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "library.h"
#include "config.h"
#include "action.h"
#include "hash.h"

#define ACT_DATELEN	(128)

typedef
struct _action_t {
	time_t		 timebuf;
	struct tm	*tm;
	char		 datebuf[ACT_DATELEN];
} action_t;

/* Cache of FILE * pointers */
static hash_t	*filehash = NULL;
#define CHECKHASH \
	do { if (!filehash) filehash = hash_new(NULL, 0); } while(0)

static void
action_print_readable(FILE *output, errmsg_t *msg, action_t *a)
{
	fprintf(output, "Date: %s\n", a->datebuf);
	fprintf(output, "PID: %d\n", msg->pid);
	fprintf(output, "Site: %s\n", msg->site);
	fprintf(output, "Port: %d\n", msg->port);
	fprintf(output, "Remote host: %s\n", msg->remote);
	fprintf(output, "Referer: %s\n", msg->referer);
	fprintf(output, "URL: %s\n", msg->url);
	fprintf(output, "Program: %s\n", msg->program);
	fprintf(output, "Facility: %s\n", msg->facility);
	fprintf(output, "Priority: %d\n", msg->priority);
	fprintf(output, "Message: %s\n", msg->message);
}

static void
action_print_machine(FILE *output, errmsg_t *msg, action_t *a)
{
	fprintf(output, "%ld:%d:%s:%d:%s:%s:%s:%s:%s:%d:%s\n",
			a->timebuf,
			msg->pid,
			msg->site,
			msg->port,
			msg->referer,
			msg->remote,
			msg->url,
			msg->program,
			msg->facility,
			msg->priority,
			msg->message);
}

static void
action_email(char *data, errmsg_t *msg, action_t *a)
{
	FILE	*output;

	output = popen(BIN_SENDMAIL " -t -i", "w");
	if (!output) {
		perror(BIN_SENDMAIL);
		return;
	}
	
	fprintf(output, "From: " ACT_MY_EMAIL "\n");
	fprintf(output, "To: %s\n", data);
	fprintf(output, "Subject: dsyslog: %s[%d]\n",
					msg->facility, msg->pid);
	fprintf(output, "\n");
	action_print_readable(output, msg, a);
	pclose(output);
}

static void
action_pipe(char *data, errmsg_t *msg, action_t *a)
{
	FILE	*output;

	CHECKHASH;

	/* No risk of collision with file because *data == '|' */
	output = hash_get(filehash, data);
	if (!output || feof(output)) {
		output = popen((data + 1), "w");
		if (!output) {
			perror(data);
			return;
		}
		hash_add(filehash, data, output);
	}

	action_print_machine(output, msg, a);
}

static void
action_exec(char *data, errmsg_t *msg, action_t *a)
{
	FILE	*output;

	output = popen((data + 1), "w");
	if (!output) {
		perror(data);
		return;
	}
	
	action_print_machine(output, msg, a);
	pclose(output);
}

static void
action_file(char *data, errmsg_t *msg, action_t *a)
{
	FILE	*output;

	CHECKHASH;

	output = hash_get(filehash, data);
	if (!output) {
		output = fopen(data, "a");
		if (!output) {
			perror(data);
			return;
		}
		hash_add(filehash, data, output);
	}

	/* As close to syslog format as I can get */
	/* XXX Remote host? */
	fprintf(output, "%s %s %s[%d]: %s (%s -> %s)\n",
					a->datebuf, msg->site, msg->facility, msg->pid,
					msg->message, msg->referer, msg->url);
	fflush(output);	/* Synchronise */
	/* We don't close the files, we cached them */
}

void
action_execute(char *data, errmsg_t *msg)
{
	action_t	 a;

	time(&a.timebuf);
	a.tm = gmtime(&a.timebuf);
	strftime(a.datebuf, (ACT_DATELEN - 1), "%b %d %H:%M:%S", a.tm);

	printf("Executing action <%s>\n", data);
	if (*data == '/')
		action_file(data, msg, &a);
	else if (*data == '|')
		action_pipe(data, msg, &a);
	else if (*data == '!')
		action_exec(data, msg, &a);
	else if (strchr(data, '@'))
		action_email(data, msg, &a);
	else
		fprintf(stderr, "Could not identify action '%s'\n", data);
}
