/*
 * Web site error daemon - message priority calculations
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <regex.h>
#include "library.h"
#include "vector.h"
#include "smalloc.h"
#include "hash.h"
#include "list.h"
#include "log.h"
#include "wregex.h"
#include "action.h"

/* Hash table object.
 * Key is a facility name
 * Data is a vector of priset_t structures, specifying actions
 *  and what priorities to execute those actions on as a bitfield
 */

#ifdef DPRINTF
#undef DPRINTF
#define DPRINTF(x) do { } while(0)
#endif

typedef
struct _priset_t {
	char		*action;
	unsigned int priorities : PRI_NPRI;
} priset_t;

/* Create a new action-priorities pair */
static priset_t *
priset_new(char *action)
{
	priset_t	*p;
	p = (priset_t *)smalloc(sizeof(priset_t));
	p->action = sstrdup(action);
	p->priorities = 0;
	return p;
}

/* Delete one */
static void
priset_delete(priset_t *p)
{
	sfree(p->action);
	sfree(p);
}

/* Compare the action fields to see if they match */
static int
priset_compare(void *a, vector_cell_t p)
{
	return strcmp((char *)(a), ((priset_t *)(p))->action);
}

/* Next... */
/* A hash table
 * key is a class or facility name
 * data is a vector of prisets
 */

/* We need to be able to do a recursive copy of the vectors.
 * Since I am not using class definitions for my objects, I shall
 * write it explicitly. Maybe soon I'll use classes.
 */

vector_t *
servicehash_vector_copy(vector_t *v)
{
	vector_t	*w;
	priset_t	*p;
	priset_t	*q;
	int			 len;
	int			 i;

	w = vector_new((vector_delete_t)priset_delete, 0);
	if (v) {	/* No harm done */
		len = vector_length(v);
		for (i = 0; i < len; i++) {
			p = vector_data(v, i);
			q = priset_new(p->action);
			q->priorities = p->priorities;
			DPRINTF((stderr, "\tCopied %s for %d\n",
							p->action, p->priorities));
			vector_push_back(w, q);
		}
	}
	return w;
}

hash_t *
servicehash_new(void)
{
	hash_t		*h;
	vector_t	*v;

	h = hash_new((hash_delete_t)vector_delete, 37);	/* Small */

	/* We want a '*' entry by default, for all services */
	v = vector_new((vector_delete_t)priset_delete, 0);
	hash_add(h, "*", v);

	return h;
}

void
servicehash_delete(hash_t *h)
{
	hash_delete(h);
}

/* Add an action (data) to a facility which is not '*' */
static void
servicehash_set_real(hash_t *h,
				char *facility,
				char *data,
				int pri,
				int action)
{
	vector_t	*v;
	priset_t	*p;
	unsigned int bit;
	int			 i;

	/* First we get the vector for the facility, else copy the '*'
	 * vector and use that */
	v = hash_get(h, facility);
	if (!v) {
		/* This function is happy with a NULL */
		DPRINTF((stderr, "Duplicating '*' entry for '%s'\n", facility));
		v = servicehash_vector_copy(hash_get(h, "*"));
		hash_add(h, facility, v);
	}

	/* Now we get a p in v by reference */
	i = vector_member(v, priset_compare, data);
	if (i == -1) {
		DPRINTF((stderr, "Creating new priset for %s\n", data));
		p = priset_new(data);
		vector_push_back(v, p);
	}
	else {
		p = vector_data(v, i);
	}

	/* Now we modify p by reference */
	bit = 1 << pri;
	DPRINTF((stderr, "Setting bit %d (%d) in '%s' (from %d)\n",
			pri, bit, p->action, p->priorities));
	if (action)
		p->priorities |= bit;
	else
		p->priorities &= ~bit;
	DPRINTF((stderr, "Priority bits are now %d\n", p->priorities));
}

void
servicehash_set(hash_t *h,
				char *facility,
				char *data,
				int pri,
				int action)
{
	vector_t	*keys;
	int			 len;
	int			 i;

	/* If we get a '*', we mean all services and the special
	 * '*' facility. We have to edit everything. */
	if (strcmp(facility, "*") == 0) {
		keys = hash_keys(h);
		len = vector_length(keys);
		for (i = 0; i < len; i++) {
			/* Yes, Steve, I'm about to clobber 'facility' */
			facility = vector_data(keys, i);
			servicehash_set_real(h, facility, data, pri, action);
		}
		vector_delete(keys);
	}
	else {
		servicehash_set_real(h, facility, data, pri, action);
	}
}

void
servicehash_execute(hash_t *h, errmsg_t *msg)
{
	vector_t	*v;
	priset_t	*p;
	int			 len;
	unsigned int bit;
	int			 i;

	v = hash_get(h, msg->facility);
	if (!v)
		v = hash_get(h, "*");
	if (!v) {
		/* NOTREACHED */
		DPRINTF((stderr, "No entry for '%s' or '*' in hash\n",
						msg->facility));
		return;
	}

	bit = 1 << msg->priority;

	len = vector_length(v);
	for (i = 0; i < len; i++) {
		p = vector_data(v, i);
		if (p->priorities & bit) {
			/* XXX Do it */
			action_execute(p->action, msg);
		}
	}
}
