/*
 * CGI email client - configuration handler
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include "library.h"
#include "cgi.h"
#include "hash.h"
#include "smalloc.h"
#include "vector.h"
#include "log.h"
#include "rope.h"

#define ERRORMSG "The '%s' field must be filled in, " \
	"please press the back button in your browser to amend the form."

/* Variables for parser */
char	*host;
char	*tpt;

void parser_parse(hash_t *cgihash, rope_t *rope, FILE *output);

static void
config_checkpresent(hash_t *cgihash, char *key)
{
	char	*data;
	data = hash_get(cgihash, key);
	if (!data) {
		log("cgiconf", PRI_CRIT, "No '%s' parameter provided - needed",
						key);
		exit(0);
	}
}

static void
config_enforce(hash_t *cgihash, char *key, char *str, rope_t *rope)
{
	char	*msg;
	char	*data;

	data = hash_get(cgihash, key);
	if ((!data) || (!strlen(data))) {
		if (str && strlen(str)) {
			msg = str;
		}
		else {
			msg = (char *)smalloc(strlen(ERRORMSG) + strlen(key) + 1);
			sprintf(msg, ERRORMSG, key);
		}
		hash_add(cgihash, "Message", msg);
		parser_parse(cgihash, rope, stdout);
		exit(0);
	}
}

static void
config_syntax(int lineno, char *str, char *msg)
{
	log("cgiconf", PRI_WARNING, "Syntax error on line %d: '%s' - %s",
					lineno, str, msg);
}

void
config_client(hash_t *cgihash)
{
	char	*thanksurl;

	thanksurl = hash_get(cgihash, "ThanksURL");
	if (thanksurl) {
		printf("Location: %s\n\n", thanksurl);
	}
	else {
		printf("Content-type: text/html\n\n"
			"Thankyou. Your message has been sent.\n");
	}
	fclose(stdout);
}

#define EMAILMSG "'%s' does not appear to be a valid email address, " \
	"please press the back button in your browser to amend the form."
void
config_email(hash_t *cgihash, char *key, char *str, rope_t *rope)
{
	char	*msg;
	char	*data;
	char	*temp;

	config_enforce(cgihash, key, str, rope);

	data = hash_get(cgihash, key);
	temp = strchr(data, '@');
	if (!temp				||
		!strchr(temp, '.')	||
		strchr(data, ',')	||
		strchr(data, '/')	) {
		if (str && strlen(str)) {
			msg = str;
		}
		else {
			msg = (char *)smalloc(strlen(EMAILMSG) + strlen(data) + 1);
			sprintf(msg, EMAILMSG, data);
		}
		hash_add(cgihash, "Message", msg);
		parser_parse(cgihash, rope, stdout);
		exit(0);
	}
}

void
config_run(hash_t *cgihash, char *hostname, char *template,
				rope_t *mrope)
{
	rope_t	*r;
	FILE	*input;
	char	 line[BUFSIZ];
	char	 cmd[BUFSIZ];
	char	 key[BUFSIZ];
	char	 data[BUFSIZ];
	int		 num;
	int		 lineno;

	host = hostname;
	tpt = template;

	r = rope_new(DIR_DATA "/");
	rope_add(r, hostname);
	rope_add(r, "/emailer/templates/");
	rope_add(r, template);
	rope_add(r, ".config");

	input = fopen(rope_chars(r), "r");
	if (!input) {
		log("cgiconf", PRI_CRIT, "Could not open %s : %s",
						rope_chars(r), strerror(errno));
		exit(0);
	}

	lineno = 0;
	while (fgets(line, BUFSIZ, input)) {
		line[strcspn(line, "\r\n")] = 0;
		lineno++;

		cmd[0] = 0;
		key[0] = 0;
		data[0] = 0;

		if (line[0] == '#') continue;	/* Comment */
		num = sscanf(line, "%s %s %s", cmd, key, data);
		if (num ==  -1) continue;		/* Blank line */

#define CMD_IS(x) (strcmp(cmd, x) == 0)
		if (CMD_IS("set")) {
			if (num >= 3)
				hash_add(cgihash, key, sstrdup(data));
			else
				config_syntax(lineno, line, "Usage: set <key> <value>");
		}
		else if (CMD_IS("require")) {
			if (num >= 2)
				config_enforce(cgihash, key, data, mrope);
			else
				config_syntax(lineno, line,
								"Usage: require <key> [<msg>]");
		}
		else if (CMD_IS("email")) {
			if (num >= 2) {
				config_email(cgihash, key, data, mrope);
			}
			else {
				config_syntax(lineno, line, "Usage: email <key>");
			}
		}
		else if (CMD_IS("unset")) {
			if (num >= 2) {
				/* XXX I hope this is OK */
				hash_remove(cgihash, key);
			}
			else {
				config_syntax(lineno, line, "Usage: unset <key>");
			}
		}
		else {
			config_syntax(lineno, line, "Unrecognised command");
		}
	}

	/* Let's do some manually */
	config_checkpresent(cgihash, "From");
	config_checkpresent(cgihash, "To");
	config_checkpresent(cgihash, "Subject");

	config_client(cgihash);
}
