/*
 * C utility library
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <string.h>
#include "library.h"
#include "hash.h"
#include "smalloc.h"
#include "vector.h"

static hash_value_t *freenode = NULL;

static hash_value_t *
hash_value_new(const char *key, void *data)
{
	hash_value_t *value;
	
	if (freenode) {
		value = freenode;
		freenode = freenode->next;
	}
	else {
		value = (hash_value_t *)smalloc(sizeof(hash_value_t));
	}

	value->key = sstrdup(key);
	value->data = data;
	value->next = NULL;

	return value;
}

static void
hash_value_delete(hash_value_t *v)
{
	free(v->key);
	v->next = freenode;
	freenode = v;
}

hash_t *
hash_new(hash_delete_t func, int hashsize)
{
    hash_t	*hashtable;
	size_t	 size;
    
    if (hashsize == 0)
        hashsize = HASHINITSIZE;
	size = hashsize * sizeof(hash_value_t *);
    
    hashtable = (hash_t *)smalloc(sizeof(hash_t));
    
    hashtable->data = (hash_value_t **)smalloc(size);
    hashtable->size = hashsize;
	hashtable->h_delete = func;

	memset(hashtable->data, 0, size);

    return hashtable;
}

void
hash_delete(hash_t *hash)
{
	int				 i;
	hash_value_t	*hv;
	hash_value_t	*temp;

	for (i = 0; i < hash->size; i++) {
		hv = hash->data[i];
		while (hv) {
			if (hash->h_delete)
				hash->h_delete(hv->data);
			temp = hv;
			hv = hv->next;
			hash_value_delete(temp);
		}
	}
	free(hash->data);
	free(hash);
}

static int
hash_function(hash_t *hashtable, const char *key)
{
    unsigned int hashvalue = 0;
    const char *cp;
    
    for (cp = key; *cp; cp++)
    {
        hashvalue = (hashvalue << 2) ^ (int)(*cp);
    }
    
    hashvalue %= hashtable->size;
    return hashvalue;
}

static hash_value_t *
hash_value_get(hash_t *hash, int offset, const char *key)
{
	hash_value_t	*value;
	hash_value_t	*temp = NULL;

	value = hash->data[offset];

	// DPRINTF((stderr, "Value is %p\n", value));
	while (value)
	{
		if (strcmp(value->key, key) == 0) {
			if (temp) {	/* Move to head of chain */
				// DPRINTF((stderr, "Caching %s\n", key));
				temp->next = value->next;
				value->next = hash->data[offset];
				hash->data[offset] = value;
			}
			return value;
		}

		temp = value;
		value = value->next;
	}

	return NULL;
}

static void
hash_value_add(hash_t *hashtable, int offset, hash_value_t *value)
{
    value->next = hashtable->data[offset];
    hashtable->data[offset] = value;
}

void
hash_add(hash_t *hashtable, const char *key, void *data)
{
    int				 offset;
    hash_value_t	*hashvalue;

	// DPRINTF((stderr, "Hashing '%s' value '%s'\n", key, (char*)data));
    
    offset = hash_function(hashtable, key);

	hashvalue = hash_value_get(hashtable, offset, key);
	if (hashvalue) {
		// DPRINTF((stderr, "Overwriting %s\n", key));
		if (hashtable->h_delete)
			hashtable->h_delete(hashvalue->data);
		hashvalue->data = data;
	}
	else {
		// DPRINTF((stderr, "Adding %s\n", key));
		hashvalue = hash_value_new(key, data);
		hash_value_add(hashtable, offset, hashvalue);
	}
}

void *
hash_get(hash_t *hashtable, const char *key)
{
    int				 offset;
	hash_value_t	*val;
    
    offset = hash_function(hashtable, key);

	// DPRINTF((stderr, "Offset is %d\n", offset));
    
    val = hash_value_get(hashtable, offset, key);
	if (!val) return NULL;
	return val->data;
}

bool
hash_remove(hash_t *h, char *key)
{
	int				 offset;
	hash_value_t	*val;

	offset = hash_function(h, key);
	val = hash_value_get(h, offset, key);

	if (!val) return FALSE;

	if (h->h_delete)
		h->h_delete(val->data);
	h->data[offset] = val->next;
	hash_value_delete(val);

	return TRUE;
}

vector_t *
hash_keys(hash_t *hashtable)
{
	int				 i;
	hash_value_t	*value;
	vector_t		*vector;

	vector = vector_new(NULL, 0);

	for (i = 0; i < hashtable->size; i++) {
		value = hashtable->data[i];
		while (value) {
			vector_push_back(vector, value->key);
			value = value->next;
		}
	}

	return vector;
}
