/*
 * Library test suite
 * Copyright (C) 1999 Open World Ltd
 * Written by Shevek <shevek@anarres.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "library.h"
#include "hash.h"
#include "cgi.h"
#include "rope.h"
#include "list.h"
#include "utils.h"
#include "log.h"
#include "wregex.h"
#include "smalloc.h"

#define H_ADD(x, y) hash_add(hash, x, strdup(y))
#define H_GET(x) ((char *)(hash_get(hash, x)))
#define H_DEL(x) (hash_remove(hash, x))
#define STR "||b|bakph|Park Hyatt Baku|Park Hyatt Baku|Baku,Azerbaijan|Baku|International Park|Hotel|ok|||||Sumgait Caspian Sea|"
#define STREQ(x, y) (strcmp(x, y) == 0)

int	 errors = 0;

void
error_if(int x)
{
	if (x) {
		errors++;
		putchar('X');
	}
	else {
		putchar('.');
	}
}

void
test_hash_item(hash_t *hash, char *key, char *val)
{
	char	*out = H_GET(key);
	error_if(!STREQ(val, out));
}

void
test_hash (void)
{
	hash_t	*hash;

	printf("Hash: ");

	hash = hash_new(sfree, 1);
	putchar('.');

	H_ADD("foo", "bar");
	test_hash_item(hash, "foo", "bar");
	H_ADD("abc", "abc");
	test_hash_item(hash, "abc", "abc");
	test_hash_item(hash, "foo", "bar");
	H_ADD("foo", "qux");
	test_hash_item(hash, "foo", "qux");
	H_DEL("foo");
	error_if((int)hash_get(hash, "foo"));

	putchar('\n');
}

void
test_explode(void)
{
	vector_t	*v;
	char		*s;
	char		*t;

	printf("Explode: ");
	v = explode(STR, "|");
	putchar('.');
	s = implode(v, "|");
	error_if(!STREQ(s, STR));
	t = implode(v, "|");
	error_if(!STREQ(s, STR));
	vector_delete(v);
	putchar('.');
	sfree(s);
	putchar('.');
	sfree(t);
	putchar('.');

	putchar('\n');
}

void
test_log(void)
{
	printf("Log: ");
	log("error404", 7, "Foo: %d %s", 7, "bar");
	putchar('.');

	putchar('\n');
}

void
test_vector_step(vector_t *v, char *s)
{
	char	*t;
	t = implode(v, ",");
	error_if(!STREQ(s, t));
	sfree(t);
	putchar('.');
}

void
test_vector(void)
{
	vector_t	*v;

	printf("Vector: ");
	v = vector_new(NULL, 0);
	putchar('.');
	vector_push_back(v, "0");
	test_vector_step(v, "0");
	vector_push_back(v, "1");
	test_vector_step(v, "0,1");
	vector_insert(v, "2", 1);
	test_vector_step(v, "0,2,1");
	vector_push_front(v, "3");
	test_vector_step(v, "3,0,2,1");
	vector_insert(v, "4", 2);
	test_vector_step(v, "3,0,4,2,1");
	vector_pop(v, 3);
	test_vector_step(v, "3,0,4,1");
	vector_pop_front(v);
	test_vector_step(v, "0,4,1");
	vector_pop_back(v);
	test_vector_step(v, "0,4");
	vector_remove(v, 0);
	test_vector_step(v, "4");

	vector_delete(v);
	putchar('.');

	putchar('\n');
}

void
test_wregex(void)
{
	wregex_t	*r;
	int			 i;

	printf("Wregex: ");
	r = wregex_new("f.o", 0);
	putchar('.');
	i = wregex_execute(r, "foo");
	error_if(i != 0);
	i = wregex_execute(r, "bar");
	error_if(i != 1);
	wregex_delete(r);
	putchar('.');

	putchar('\n');
}


int
main(void)
{
	library_init("test");

	test_hash();
	test_explode();
	test_log();
	test_vector();
	test_wregex();

	if (errors) {
		printf("ERROR: Test failed. %d errors.\n", errors);
		exit(1);
	}
	else {
		printf("Program completed succesfully.\n");
	}

	return 0;
}
